package edu.umd.dawn.common.exceptions;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Map;
import java.util.Set;

import lombok.extern.slf4j.Slf4j;
import org.slf4j.MDC;
import org.springframework.beans.ConversionNotSupportedException;
import org.springframework.beans.TypeMismatchException;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.ResponseEntity;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.http.converter.HttpMessageNotWritableException;
import org.springframework.stereotype.Component;
import org.springframework.validation.BindException;
import org.springframework.validation.FieldError;
import org.springframework.web.HttpMediaTypeNotAcceptableException;
import org.springframework.web.HttpMediaTypeNotSupportedException;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.MissingPathVariableException;
import org.springframework.web.bind.MissingServletRequestParameterException;
import org.springframework.web.bind.ServletRequestBindingException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.context.request.async.AsyncRequestTimeoutException;
import org.springframework.web.multipart.support.MissingServletRequestPartException;
import org.springframework.web.servlet.NoHandlerFoundException;

@ControllerAdvice
@Component
@Slf4j
public class CustomExceptionHandler {

    private final Map<Class<?>, DawnExceptionParameters> MVC_EXCEPTION_MAP = Map.ofEntries(
        Map.entry(HttpRequestMethodNotSupportedException.class, BaseExceptions.METHOD_NOT_ALLOWED),
        Map.entry(HttpMediaTypeNotSupportedException.class, BaseExceptions.UNSUPPORTED_MEDIA_TYPE),
        Map.entry(HttpMediaTypeNotAcceptableException.class, BaseExceptions.NOT_ACCEPTABLE),
        Map.entry(MissingPathVariableException.class, BaseExceptions.INTERNAL_SERVER_ERROR),
        Map.entry(MissingServletRequestParameterException.class, BaseExceptions.BAD_REQUEST),
        Map.entry(ServletRequestBindingException.class, BaseExceptions.BAD_REQUEST),
        Map.entry(ConversionNotSupportedException.class, BaseExceptions.INTERNAL_SERVER_ERROR),
        Map.entry(TypeMismatchException.class, BaseExceptions.BAD_REQUEST),
        Map.entry(HttpMessageNotReadableException.class, BaseExceptions.BAD_REQUEST),
        Map.entry(HttpMessageNotWritableException.class, BaseExceptions.INTERNAL_SERVER_ERROR),
        Map.entry(MethodArgumentNotValidException.class, BaseExceptions.BAD_REQUEST),
        Map.entry(MissingServletRequestPartException.class, BaseExceptions.BAD_REQUEST),
        Map.entry(NoHandlerFoundException.class, BaseExceptions.NOT_FOUND),
        Map.entry(AsyncRequestTimeoutException.class, BaseExceptions.SERVICE_UNAVAILABLE)
    );

    @Value("${config.serviceName}")
    private String source;

    private String getStackTrace(Exception exception) {
        StringWriter sw = new StringWriter();
        exception.printStackTrace(new PrintWriter(sw));
        String exceptionAsString = sw.toString();
        return exceptionAsString;
    }

    private void logException(DawnException exception) {
        if (exception.getBase() != null) {
            MDC.put("stackTrace", getStackTrace(exception.getBase()));
            MDC.put("exceptionMessage", exception.getBase().getMessage());
        } else {
            MDC.put("stackTrace", getStackTrace(exception));
            MDC.put("exceptionMessage", exception.getMessage());
        }
        log.error(exception.getMessage());
        MDC.remove("stackTrace");
        MDC.remove("exceptionMessage");
    }

    private ResponseEntity<Object> returnDawnException(DawnException exception) {
        logException(exception);
        return ExceptionResponse.fromDawnException(exception, source).toResponse();
    }

    private ResponseEntity<Object> wrapExceptionAs(RuntimeException exception, DawnExceptionParameters parameters) {
        DawnException wrapped = new DawnException(parameters, exception);
        return returnDawnException(wrapped);
    }

    private ResponseEntity<Object> wrapExceptionAs(Exception exception, DawnExceptionParameters parameters) {
        DawnException wrapped = new DawnException(parameters, exception);
        return returnDawnException(wrapped);
    }

    @ExceptionHandler(BindException.class)
    protected ResponseEntity<Object> handleBindException(BindException ex, WebRequest request) {

        FieldError err = ex.getFieldError();
        DawnException wrapped = new DawnException(
                BaseExceptions.BAD_REQUEST,
                ex,
                "Value " + err.getRejectedValue() + " is invalid for field " + err.getField());

        return returnDawnException(wrapped);
    }

    @ExceptionHandler(ArrayIndexOutOfBoundsException.class)
    public ResponseEntity<Object> handleArrayIndexOutOfBoundsException(
            ArrayIndexOutOfBoundsException exception, WebRequest webRequest) {
        return wrapExceptionAs(exception, BaseExceptions.OUT_OF_BOUNDS);
    }
    
    @ExceptionHandler(DawnException.class)
    public ResponseEntity<Object> handleDawnExceptions(DawnException exception, WebRequest webRequest) {
        return returnDawnException(exception);
    }

    // begin common spring mvc exceptions
    @ExceptionHandler({
        HttpRequestMethodNotSupportedException.class,
        HttpMediaTypeNotSupportedException.class,
        HttpMediaTypeNotAcceptableException.class,
        MissingPathVariableException.class,
        MissingServletRequestParameterException.class,
        ServletRequestBindingException.class,
        ConversionNotSupportedException.class,
        TypeMismatchException.class,
        HttpMessageNotReadableException.class,
        HttpMessageNotWritableException.class,
        MethodArgumentNotValidException.class,
        MissingServletRequestPartException.class,
        NoHandlerFoundException.class,
        AsyncRequestTimeoutException.class
    })
    public ResponseEntity<Object> handleMvcError(Exception exception, WebRequest webRequest) {
        Set<Class<?>> classList = MVC_EXCEPTION_MAP.keySet();
        for (Class<?> clazz : classList) {
            if (clazz.isInstance(exception)) {
                return wrapExceptionAs(exception, MVC_EXCEPTION_MAP.get(clazz));
            }
        }
        return wrapExceptionAs(exception, BaseExceptions.UNHANDLED_INTERNAL_SERVER_ERROR);
    }
    // end common spring mvc exceptions

    @ExceptionHandler(RuntimeException.class)
    public ResponseEntity<Object> handleUnHandledException(RuntimeException exception, WebRequest webRequest) {
        return wrapExceptionAs(exception, BaseExceptions.UNHANDLED_INTERNAL_SERVER_ERROR);
    }

    @ExceptionHandler(Exception.class)
    public ResponseEntity<Object> handleUnHandledException(Exception exception, WebRequest webRequest) {
        return wrapExceptionAs(exception, BaseExceptions.UNHANDLED_INTERNAL_SERVER_ERROR);
    }

    @ExceptionHandler(Throwable.class)
    public ResponseEntity<Object> handleUnHandledException(Throwable exception, WebRequest webRequest) {
        return wrapExceptionAs((Exception)exception, BaseExceptions.UNHANDLED_INTERNAL_SERVER_ERROR);
    }
}
